/*
 *  integral_function.c
 *  Astronomy
 *
 *  Created by Travis Desell on 2/21/07.
 *  Copyright 2007 __MyCompanyName__. All rights reserved.
 *
 */

/****
        *       BOINC includes
*****/

#ifdef _WIN32
#include "boinc_win.h"
#else
#include "config.h"
#endif

#ifndef _WIN32
#include <cstdio>
#include <cctype>
#include <cstring>
#include <cstdlib>
#include <csignal>
#endif

//#define BOINC_APP_GRAPHICS

#ifdef BOINC_APP_GRAPHICS
#include "graphics_api.h"
#include "graphics_lib.h"
#endif

#include "diagnostics.h"
#include "str_util.h"
#include "util.h"
#include "filesys.h"
#include "boinc_api.h"
#include "mfile.h"

using std::string;

#define INTEGRAL_CHECKPOINT_FILE "astronomy_integral_checkpoint"

/****
	*	Astronomy includes
*****/
#include "integral_function.h"
#include "parameters.h"
#include "volume.h"
#include "probability.h"
#include "stCoords.h"
#include "atSurveyGeometry.h"
#include "star_points.h"

#ifndef _WIN32
#define pi M_PI
#else
#define pi 3.14159265358979323846
#endif
#define deg (180.0/pi)

bool convolve = false;
double integral_time;

double background_integral;
double *stream_integrals;

int mu_step_current, nu_step_current, r_step_current;

double get_background_integral() {
	return background_integral;
}

double* get_stream_integrals() {
	return stream_integrals;
}

double get_integral_time() {
	return integral_time;
}

void load_integral_checkpoint() {
	FILE* checkpoint;
	char chkpt_path[512];

	fprintf(stderr,"APP: astronomy reading integral checkpoint file\n");

	// See if there's a valid checkpoint file.
	// If so seek input file
	boinc_resolve_filename(INTEGRAL_CHECKPOINT_FILE, chkpt_path, sizeof(chkpt_path));
	checkpoint = boinc_fopen(chkpt_path, "r");
	if (checkpoint) {
		//double likelihood;
		fscanf(checkpoint, "mu_step_current: %d, nu_step_current: %d, r_step_current: %d, background_integral: %lf\n", &mu_step_current, &nu_step_current, &r_step_current, &background_integral);
		stream_integrals = (double*)malloc(sizeof(double) * get_number_streams());
		read_double_array(checkpoint, "stream_integrals: ", get_number_streams(), &stream_integrals);
		fclose(checkpoint);
	} else {
		mu_step_current = 0;
		nu_step_current = 0;
		r_step_current = 0;
		background_integral = 0;
		stream_integrals = (double*)malloc(sizeof(double) * get_number_streams());
		for (int i = 0; i < get_number_streams(); i++) {
			stream_integrals[i] = 0;
		}
	}

	fprintf(stderr,"APP: astronomy read integral checkpoint finished\n");
}

int do_integral_checkpoint() {
	string resolved_name;
	int retval;

	FILE* checkpoint = fopen("temp", "w");

	if (checkpoint) {
		fprintf(checkpoint, "mu_step_current: %d, nu_step_current: %d, r_step_current: %d, background_integral: %lf\n", mu_step_current, nu_step_current, r_step_current, background_integral); 
		print_double_array(checkpoint, "stream_integrals: ", get_number_streams(), stream_integrals);
		fclose(checkpoint);
		//return 0;
	} else {
		fprintf(stderr, "APP: astronomy ERROR GENERATING CHECKPOINT FILE\n");
		return 1;
	}

	fprintf(stderr,"APP: astronomy integral checkpointing\n");

	boinc_resolve_filename_s(INTEGRAL_CHECKPOINT_FILE, resolved_name);
	retval = boinc_rename("temp",resolved_name.c_str());
	if(retval) return retval;

	fprintf(stderr,"APP: astronomy integral checkpoint done\n");
	return 0;
}

void calculate_integrals() {
	int wedge;
	double background_weight;
	double *background_parameters;
	double *stream_weights;
	double **stream_parameters;

	double nu, nu_min, nu_max, nu_step, nu_steps;
	double mu, mu_min, mu_max, mu_step, mu_steps;
	double r, r_min, r_max, r_step, r_steps, log_r, next_r;

	int num_calls;
	double total_volume, V;
	double ir, id;
	double ra, dec, point0, point1;
	double integral_point[3], xyz[3];

	int retval;
	double f;
	bool first_run = true;

	wedge = get_wedge();
	background_weight = get_background_weight();
	background_parameters = get_background_parameters();
	stream_weights = get_stream_weights();
	stream_parameters = get_stream_parameters();

	nu_min = get_nu_min();
	nu_max = get_nu_max();
	nu_step = get_nu_step();
	nu_steps = get_nu_steps();
	
	mu_min = get_mu_min();
	mu_max = get_mu_max();
	mu_step = get_mu_step();
	mu_steps = get_mu_steps();
	
	r_min = get_r_min();
	r_max = get_r_max();
	r_step = get_r_step();
	r_steps = get_r_steps();

	background_integral = 0;
	num_calls = 0;
	V = 0;
	total_volume = 0;

	//printf("r_min: %lf, r_max: %lf, r_step: %lf, r_steps: %lf\n", r_min, r_max, r_step, r_steps);
	//printf("mu_min: %lf, mu_max: %lf, mu_step: %lf, mu_steps: %lf\n", mu_min, mu_max, mu_step, mu_steps);
	//printf("nu_min: %lf, nu_max: %lf, nu_step: %lf, nu_steps: %lf\n", nu_min, nu_max, nu_step, nu_steps);
	//printf("wedge: %d\n", wedge);

	//printf("background_weight: %lf\n", background_weight);
	//print_double_array(stdout, "background_parameters: ", 4, background_parameters);
	//printf("stream_weight: %lf\n", stream_weights[0]);
	//print_double_array(stdout, "stream_parameters: ", 5, stream_parameters[0]);
	//printf("\n");

	load_integral_checkpoint();

	for ( ; mu_step_current < mu_steps; mu_step_current++) {
		mu = mu_min + mu_step_current*mu_step;

		if(!first_run) nu_step_current=0;
		for ( ; nu_step_current < nu_steps; nu_step_current++) {
			nu = nu_min + nu_step_current*nu_step;

			if(!first_run) r_step_current=0;
			for ( ; r_step_current < r_steps; r_step_current++) {
				log_r = r_min + r_step_current*r_step;

				r = pow(10.0, (log_r-14.2)/5.0);
				next_r = pow(10.0, (log_r+r_step-14.2)/5.0);

				if (wedge > 0) {
					ir = (pow(next_r,3.0) - pow(r, 3.0))/3.0;
					id = cos((90 - nu - nu_step)/deg) - cos((90 - nu)/deg);

					V = ir * id * mu_step / deg;

					ra = 0.0;
					dec = 0.0;
					point0 = 0.0;
					point1 = 0.0;

					atGCToEq(mu + 0.5 * mu_step, nu + 0.5 * nu_step, &ra, &dec, get_node(), wedge_incl(wedge));
					atEqToGal(ra, dec, &point0, &point1);
					integral_point[0] = point0;
					integral_point[1] = point1;
					integral_point[2] = (next_r+r)/2.0;

					total_volume += V;
				} else {
					V = mu_step * nu_step * r_step;
					xyz[0] = mu + 0.5*mu_step;
					xyz[1] = nu + 0.5*nu_step;
					xyz[2] = r + 0.5*r_step;
					xyz2lbr(xyz, integral_point);
				}

				double bg_prob = 0.0;
				double st_prob = 0.0;

				if (convolve) {
//					bg_prob = Probability.stPbxConvolved(integral_point, background_parameters);
//					st_prob = Probability.stPsgConvolved(integral_point, stream_parameters[0], wedge);
				} else {
					bg_prob = stPbx(integral_point, background_parameters);
					st_prob = stPsg(integral_point, stream_parameters[0]);
				}

				background_integral += bg_prob * V;
				stream_integrals[0] += st_prob * V;
				first_run = false;
				num_calls++;

				/*if (num_calls%100000 == 0) {
					fprintf(stderr, "integral point: %lf %lf %lf\n", integral_point[0], integral_point[1], integral_point[2]);
					fprintf(stderr, "ir: %lf, id: %lf, deg: %lf\n", ir, id, deg);
					fprintf(stderr, "V: %lf, bg_prob: %lf, st_prob: %lf\n", V, bg_prob, st_prob);
					fprintf(stderr, "bg_prob: %lf, st_prob: %lf, total_volume: %lf\n", bg_prob, st_prob, total_volume);
					fprintf(stderr, "completed %d calls. background_integral: %lf, stream_integral[0]: %lf\n", num_calls, background_integral, stream_integral[0]);
					}*/

				if(boinc_time_to_checkpoint()) {
				        retval = do_integral_checkpoint();

					if (retval) {
					          fprintf(stderr,"APP: astronomy checkpoint failed %d\n",retval);
						  exit(retval);
					}
					boinc_checkpoint_completed();
				}

				f = mu_step_current*nu_steps*r_steps + nu_step_current*r_steps + r_step_current;
				f /= mu_steps*nu_steps*r_steps + get_number_streams()*(get_number_star_points()+.001)/100;
				//if(num_calls%100000 == 0) printf("%f percent done\n",f*100);
				boinc_fraction_done(f);
			}
			r_step_current = 0;
		}
		nu_step_current = 0;
	}

	do_integral_checkpoint();
}
