/*
 *  likelihood_function.c
 *  Astronomy
 *
 *  Created by Travis Desell on 2/21/07.
 *  Copyright 2007 __MyCompanyName__. All rights reserved.
 *
 */

/********
        *     Include files for BOINC
********/
#ifdef _WIN32
#include "boinc_win.h"
#else
#include "config.h"
#endif

#ifndef _WIN32
#include <cstdio>
#include <cctype>
#include <cstring>
#include <cstdlib>
#include <csignal>
#include <unistd.h>
#include <fcntl.h>
#endif

//#define BOINC_APP_GRAPHICS

#ifdef BOINC_APP_GRAPHICS
#include "graphics_api.h"
#include "graphics_lib.h"
#endif

#include "diagnostics.h"
#include "str_util.h"
#include "util.h"
#include "filesys.h"
#include "boinc_api.h"
#include "mfile.h"

using std::string;

#define LIKELIHOOD_CHECKPOINT_FILE "astronomy_likelihood_checkpoint"

/********
        *     Include files for astronomy
********/

#include "integral_function.h"
#include "likelihood_function.h"
#include "star_points.h"
#include "volume.h"
#include "parameters.h"
#include "probability.h"

int current_star_point;
int current_stream;
double prob_sum;
double star_prob;
double sum_integrals;

void load_likelihood_checkpoint() {
	FILE* checkpoint;
	char chkpt_path[512];

	fprintf(stderr,"APP: astronomy reading likelihood checkpoint file\n");

	// See if there's a valid checkpoint file.
	// If so seek input file
        boinc_resolve_filename(LIKELIHOOD_CHECKPOINT_FILE, chkpt_path, sizeof(chkpt_path));
	checkpoint = boinc_fopen(chkpt_path, "r");
	if (checkpoint) {
		fscanf(checkpoint, "current_star_point: %d, current_stream: %d, star_prob: %lf, prob_sum: %lf, sum_integrals: %lf\n", &current_star_point, &current_stream, &star_prob, &prob_sum, &sum_integrals);
		fclose(checkpoint);
	} else {
		current_star_point = 0;
		current_stream = 0;
		prob_sum = 0;
		star_prob = 0;
		sum_integrals = 0;
	}
	fprintf(stderr,"APP: astronomy read likelihood checkpoint finished\n");
}

int do_likelihood_checkpoint() {
	string resolved_name;
	int retval;

	FILE* checkpoint = fopen("temp", "w");

	if (checkpoint) {
		fprintf(checkpoint, "current_star_point: %d, current_stream: %d, star_prob: %lf, prob_sum: %lf, sum_integrals: %lf\n", current_star_point, current_stream, star_prob, prob_sum, sum_integrals);
		fclose(checkpoint);
		//return 0;
	} else {
		fprintf(stderr,"APP: astronomy ERROR GENERATING CHECKPOINT FILE\n");
		return 1;
	}

	fprintf(stderr,"APP: astronomy likelihood checkpointing\n");

	boinc_resolve_filename_s(LIKELIHOOD_CHECKPOINT_FILE, resolved_name);
	retval = boinc_rename("temp",resolved_name.c_str());
	if(retval) return retval;

	fprintf(stderr,"APP: astronomy likelihood checkpoint done\n");
	return 0;
}

double calculate_likelihood(double background_integral, double *stream_integrals) {
	int wedge;
	double background_weight;
	double *background_parameters;
	double *stream_weights, *exp_stream_weights;
	double **stream_parameters;
	double **star_points;
	double sum_exp_weights;
	double *star_coords;
	int num_zero, bad_jacobians;
	int i=0;

	bool new_formula, convolve;
	new_formula = false;
	convolve = false;

	int retval, num_calls=0;
	bool first_run = true;
	double f;

	wedge = get_wedge();
	background_weight = get_background_weight();
	background_parameters = get_background_parameters();
	stream_weights = get_stream_weights();
	stream_parameters = get_stream_parameters();
	star_points = get_star_points();

	background_weight = exp(background_weight);
	sum_exp_weights = 0.0;
	exp_stream_weights = (double*)malloc(sizeof(double) * get_number_streams());

	for (; i < get_number_streams(); i++) {
		exp_stream_weights[i] = exp(stream_weights[i]);
		sum_exp_weights += exp(stream_weights[i]);
	}
	//update: added background_weight to sum_exp_weights
	sum_exp_weights += background_weight;

	prob_sum = 0.0;
	num_zero = 0;
	bad_jacobians = 0;   

	load_likelihood_checkpoint();

	for (; current_star_point < get_number_star_points(); current_star_point++) {
		star_coords = star_points[current_star_point];

		if (current_stream == 0) {
			star_prob = 0.0;
			sum_integrals = 0.0;
		}

		if(!first_run) current_stream = 0;
		for (; current_stream < get_number_streams(); current_stream++) {
			if (new_formula) {
				//update: added wedge arg to stPsg call
				if (convolve) ;//star_prob += stPsgConvolved(star_coords, stream_parameters[current_stream], wedge) * stream_weights[current_stream];
				else star_prob += stPsg(star_coords, stream_parameters[current_stream]) * exp_stream_weights[current_stream];
				sum_integrals += exp_stream_weights[current_stream] * stream_integrals[current_stream];
			} else {
				//update: added wedge arg to stPsg call
				if (convolve) ;//star_prob += (stPsgConvolved(star_coords, stream_parameters[current_stream], wedge)/stream_integrals[current_stream]) * stream_weights[current_stream];
				else star_prob += (stPsg(star_coords, stream_parameters[current_stream])/stream_integrals[current_stream]) * exp_stream_weights[current_stream];
			}
			first_run = false;
			num_calls++;

			if (boinc_time_to_checkpoint()) {
		                retval = do_likelihood_checkpoint();
			        if(retval) {
			                fprintf(stderr, "APP: astronomy checkpoint failed %d\n",retval);
				        exit(retval);
			        }
					boinc_checkpoint_completed();
		        }

			f = get_mu_steps()*get_nu_steps()*get_r_steps() + (get_number_streams())/100000 + (current_star_point*get_number_streams() + current_stream)/100;
			f /= get_mu_steps()*get_nu_steps()*get_r_steps() + get_number_streams()*(get_number_star_points()+.001)/100;
			//if(num_calls%100000 == 0) printf("%f percent done\n",f*100);
			boinc_fraction_done(f);
		}
		current_stream = 0;

		//Find the background probability
		if (new_formula) {
			if (convolve) ;//star_prob += stPbxConvolved(star_coords, background_parameters) * background_weight;
			else star_prob += stPbx(star_coords, background_parameters) * background_weight;
			sum_integrals += background_weight * background_integral;
		} else {
			if (convolve) ;//star_prob += (stPbxConvolved(star_coords, background_parameters)/background_integral) * background_weight;
			else star_prob += (stPbx(star_coords, background_parameters)/background_integral) * background_weight;
		}

//		fprintf(stderr, "star_prob: %lf, sum_exp_weights: %lf, background_weight: %lf\n", star_prob, sum_exp_weights, background_weight);

		//Calculate the probability for this star.
		star_prob /= sum_exp_weights;
		if (new_formula) star_prob /= sum_integrals;		
		
		//update: check star_prob==0, not prob_sum
		if (star_prob != 0.0) {
			prob_sum += log(star_prob)/log(10.0);
		} else {
			num_zero++;
			prob_sum -= 238.0;
		}

	}
	do_likelihood_checkpoint();

	prob_sum /= (get_number_star_points()-bad_jacobians);

	free(exp_stream_weights);

	return prob_sum;
}


