#include "config.h"
#include <cstdio>

#include "boinc_db.h"
#include "backend_lib.h"
#include "util.h"
#include "sched_config.h"
#include "sched_msgs.h"
#include "sched_util.h"
#include "assimilate_handler.h"
#include "validate_util.h"
#include "string.h"

#include "parameters.h"
#include "boinc_wu_generator.h"
#include "boinc_genetic_search.h"

using std::vector;
using std::string;

/*
 *	work_buffer is the number of unsent work units we should have
 *	generated at all times.
 */
const int work_buffer = 100;

/*
 *	This gets the type of the search from the name of the work unit
 *	I'm thinking the work unit name could be:
 *		<search_type>_<search_number>_<wu_number>
 */
void get_search_type(char* wu_name, char* search_type) {
	unsigned int i;
	for (i = 0; i < strlen(wu_name); i++) if (wu_name[i] == '_') break;
	if (i >= strlen(wu_name)) sprintf(search_type, "unknown");
	strncpy(search_type, wu_name, sizeof(char)*i);
	search_type[i] = '\0';
}

/*
 *	This gets the id of the search the work unit is from, from the name of the work unit
 *	I'm thinking that the search_id could be:
 *		<search_type>_<search_number>
 *	from the work unit name.
 */
void get_search_id(char* wu_name, char* search_id) {
	unsigned int i;
	for (i = 0; i < strlen(wu_name); i++) if (wu_name[i] == '_') break;
	i++;
	for (; i < strlen(wu_name); i++) if (wu_name[i] == '_') break;
	if (i >= strlen(wu_name)) sprintf(search_id, "unknown");
	strncpy(search_id, wu_name, sizeof(char)*i);
	search_id[i] = '\0';
}

/*
 *	A general handler for multiple types of searches
 */
int assimilate_handler(WORKUNIT& wu, vector<RESULT>& /*results*/, RESULT& canonical_result) {
	SCOPE_MSG_LOG scope_messages(log_messages, SCHED_MSG_LOG::MSG_NORMAL);
	scope_messages.printf("[%s] Assimilating\n", wu.name);

	if (wu.canonical_resultid) {
		string output_file_name;

		scope_messages.printf("[%s] Found canonical result\n", wu.name);
		log_messages.printf_multiline(SCHED_MSG_LOG::MSG_DEBUG, canonical_result.xml_doc_out, "[%s] canonical result", wu.name);

		if (!(get_output_file_path(canonical_result, output_file_name))) {
			scope_messages.printf("[%s] Output file path %s\n", wu.name, output_file_name.c_str());
		}

		char search_type[3];
		get_search_type(wu.name, search_type);
		char search_id[256];
		get_search_id(wu.name, search_id);

		if (strcmp(search_type, "gs")) {
			/*
			 *	The work unit was from a genetic search, pass the result and the search
			 *	id to the genetic search handler.
			 */
			scope_messages.printf("[%s] Updating genetic search.\n", wu.name);

			/*
			 *      Read the likelihood and the parameters that generated the result from the work unit.
			 */
			scope_messages.printf("[%s] assimiating, found canonical result, output file path: %s\n", wu.name, output_file_name.c_str());
			FILE* result_file = fopen(output_file_name.c_str(), "r");
			double fitness;
			double *parameters_used;
			int number_parameters, i;
			fscanf(result_file, "calculated likelihood: %lf\n", &fitness);
			fscanf(result_file, "parameters: [%d]:", &number_parameters);
			parameters_used = (double*)malloc(sizeof(double) * number_parameters);
			for (i = 0; i < number_parameters; i++) {
				fscanf(result_file, " %lf", &parameters_used[i]);
			}
			fclose(result_file);
			gs__process_workunit(search_id, parameters_used, fitness);
			free(parameters_used);

			/*
			 *	Determine the number of work units that should be generate, and
			 *	tell the genetic search to make parameters and invoke add_work_unit
			 *	with them.
			 */
			int unsent_workunits;
			count_unsent_results(unsent_workunits, 0);

			if (unsent_workunits < work_buffer) {
				gs__generate_workunits(search_id, add_work_unit);
			}
		} else {
			scope_messages.printf("[%s] Unknown search: %s\n", wu.name, search_type);
		}
	} else {
		scope_messages.printf("[%s] No canonical result\n", wu.name);
	}

	if (wu.error_mask&WU_ERROR_COULDNT_SEND_RESULT)		log_messages.printf(SCHED_MSG_LOG::MSG_CRITICAL, "[%s] Error: couldn't send a result\n", wu.name);
	if (wu.error_mask&WU_ERROR_TOO_MANY_ERROR_RESULTS)	log_messages.printf(SCHED_MSG_LOG::MSG_CRITICAL, "[%s] Error: too many error results\n", wu.name);
	if (wu.error_mask&WU_ERROR_TOO_MANY_TOTAL_RESULTS)	log_messages.printf(SCHED_MSG_LOG::MSG_CRITICAL, "[%s] Error: too many total results\n", wu.name);
	if (wu.error_mask&WU_ERROR_TOO_MANY_SUCCESS_RESULTS)	log_messages.printf(SCHED_MSG_LOG::MSG_CRITICAL, "[%s] Error: too many success results\n", wu.name);
	return 0;
}

const char *BOINC_RCSID_8f6a5a2d27 = "$Id: assimilator_astronomy.C,v 1.7 2007/10/25 08:58:39 deselt Exp $";
