/********
	*	This is a skeleton for how all GML/MPI code should look.
	*	the gml directory should be in your $PATH file.
 ********/

/********
	*	Include the function evaluator
 ********/
#include	"../framework/gml.h"
#include	"../framework/evaluator.h"

/********
	*	Includes for the different search methods
 ********/
#include	"../framework/gradient_descent.h"
#include	"../framework/genetic_search.h"


/********
	*	Includes for astronomy
 ********/
#include "parameters.h"
#include "volume.h"
#include "star_points.h"
#include "integral_function.h"
#include "likelihood_function.h"


int integral_parameter_length;
int integral_results_length;
int likelihood_parameter_length;
int likelihood_results_length;


void read_data(int rank, int max_rank) {
	read_likelihood_data(rank, max_rank);
	read_integral_data(rank, max_rank);
}

double* astronomy_integral_function(double* parameters) {
	double background_integral;
	double* stream_integrals;
	double* results;
	int i;

	set_parameters_from_double_array(parameters);

	calculate_integrals();
	background_integral = get_background_integral();
	stream_integrals = get_stream_integrals();

	results = (double*) malloc( sizeof(double) * integral_results_length );

	results[0] = background_integral;
	for (i = 0; i < get_number_streams(); i++) {
		results[i+1] = stream_integrals[i];
	}

	return results;
}

double* astronomy_integral_combinator(double* integral_results, int number_slices) {
	double* combined_integral_results;
	int i, j;

	combined_integral_results = (double*)malloc(sizeof(double) * integral_results_length);
	for (i = 0; i < integral_results_length; i++) {
		combined_integral_results[i] = 0;
	}

	for (i = 0; i < number_slices; i++) {
		for (j = 0; j < integral_results_length; j++) {
			combined_integral_results[j] += integral_results[(i*integral_results_length) + j];
		}
	}

	return combined_integral_results;
}

double* astronomy_likelihood_function(double* likelihood_parameters) {
	double* likelihood_result;
	int i;
	double background_integral;
	double* stream_integrals;

	background_integral = likelihood_parameters[0];
	stream_integrals = (double*)malloc(sizeof(double) * get_number_streams());
	for (i = 0; i < get_number_streams(); i++) {
		stream_integrals[i] = likelihood_parameters[i+1];
	}

	likelihood_result = (double*)malloc(sizeof(double));
	likelihood_result[0] = calculate_likelihood(background_integral, stream_integrals);
	return likelihood_result;
}

double astronomy_likelihood_combinator(double* integral_results, int number_slices) {
	double result;
	int i;

	result = 0;
	for (i = 0; i < number_slices; i++) {
		result += integral_results[i];
	}

	return -(result/get_total_star_points());
}


int main(int number_arguments, char** arguments) {
	double* initial_parameters;
	int parameter_length;
	double* step_sizes;
	int step_size_length;
	int i;
	int max_evaluations;

	srand48(1000);

	max_evaluations = 50000;
	evaluator__max_evaluations(max_evaluations);

	read_parameters();
	get_parameters_as_double_array(&initial_parameters, &parameter_length);
	get_step_sizes_as_double_array(&step_sizes, &step_size_length);

	/********
		*	Distribute data
	 ********/
	evaluator__init( number_arguments, arguments );
	evaluator__init_data( read_data );

	/********
		*	Initialize the evaluator with the integral and likelihood functions
	 ********/
	integral_parameter_length = parameter_length;
	integral_results_length = get_number_streams() + 1;
	likelihood_parameter_length = get_number_streams() + 1;
	likelihood_results_length = 1;

	evaluator__init_integral(astronomy_integral_function, integral_parameter_length, astronomy_integral_combinator, integral_results_length);
	evaluator__init_likelihood(astronomy_likelihood_function, likelihood_parameter_length, astronomy_likelihood_combinator, likelihood_results_length);

	/********
		*	Run the search method.
	 ********/
//	for (i = 0; i < 500; i++) {
//		printf("\n\nCGD ITERATION: %d\n", i);
//		get_random_parameters(&initial_parameters, &parameter_length);
//		conjugate_gradient_descent( initial_parameters, step_sizes, parameter_length );
//	}

	genetic_search__probabilistic(get_parameter_min(), get_parameter_max(), parameter_length, 300, 0.8, max_evaluations);
}
