/*
 *  integral_function.c
 *  Astronomy
 *
 *  Created by Travis Desell on 2/21/07.
 *  Copyright 2007 __MyCompanyName__. All rights reserved.
 *
 */

#include "mpi.h"

/****
	*	Astronomy includes
*****/
#include "integral_function.h"
#include "parameters.h"
#include "volume.h"
#include "probability.h"
#include "stCoords.h"
#include "atSurveyGeometry.h"

#define pi M_PI
#define deg (180.0/pi)

bool convolve = false;
double integral_time;

double background_integral;
double *stream_integrals;

double get_background_integral() {
	return background_integral;
}

double* get_stream_integrals() {
	return stream_integrals;
}

double get_integral_time() {
	return integral_time;
}

void calculate_integrals() {
	int wedge;
	double background_weight;
	double *background_parameters;
	double *stream_weights;
	double **stream_parameters;

	double nu, nu_min, nu_max, nu_step, nu_steps;
	double mu, mu_min, mu_max, mu_step, mu_steps;
	double r, r_min, r_max, r_step, r_steps, log_r, next_r;

	int num_calls;
	double total_volume, V;
	double ir, id;
	double ra, dec, point0, point1;
	double integral_point[3], xyz[3], stream_integral;
	int i, j, k;

	double start, finish;

	start = MPI_Wtime();

	wedge = get_wedge();
	background_weight = get_background_weight();
	background_parameters = get_background_parameters();
	stream_weights = get_stream_weights();
	stream_parameters = get_stream_parameters();

	nu_min = get_nu_min();
	nu_max = get_nu_max();
	nu_step = get_nu_step();
	nu_steps = get_nu_steps();
	
	mu_min = get_mu_min();
	mu_max = get_mu_max();
	mu_step = get_mu_step();
	mu_steps = get_mu_steps();
	
	r_min = get_r_min();
	r_max = get_r_max();
	r_step = get_r_step();
	r_steps = get_r_steps();

	background_integral = 0;
	stream_integral = 0;
	num_calls = 0;
	V = 0;
	total_volume = 0;

//	fprintf(stderr, "r_min: %lf, r_max: %lf, r_step: %lf, r_steps: %lf\n", r_min, r_max, r_step, r_steps);
//	fprintf(stderr, "mu_min: %lf, mu_max: %lf, mu_step: %lf, mu_steps: %lf\n", mu_min, mu_max, mu_step, nu_steps);
//	fprintf(stderr, "nu_min: %lf, nu_max: %lf, nu_step: %lf, nu_steps: %lf\n", nu_min, nu_max, nu_step, nu_steps);
//	fprintf(stderr, "wedge: %d\n", wedge);

//	fprintf(stderr, "background_weight: %lf\n", background_weight);
//	print_double_array(stderr, "background_parameters: ", 4, background_parameters);
//	fprintf(stderr, "stream_weight: %lf\n", stream_weights[0]);
//	print_double_array(stderr, "stream_parameters: ", 5, stream_parameters[0]);
//	fprintf(stderr, "\n");

	stream_integrals = (double*)malloc(sizeof(double)*1);
	
	for (i = 0; i < mu_steps; i++) {
		mu = mu_min + i*mu_step;
//		printf("mu: %lf, steps %d of %lf\n", mu, i, mu_steps);
		for (j = 0; j < nu_steps; j++) {
			nu = nu_min + j*nu_step;
//			printf("nu: %lf, steps %d of %ld\n", nu, i, nu_steps);
			for (k = 0; k < r_steps; k++) {
				log_r = r_min + k*r_step;

				r = pow(10.0, (log_r-14.2)/5.0);
				next_r = pow(10.0, (log_r+r_step-14.2)/5.0);

//				printf("r: %lf, log_r: %lf, next_r: %lf, steps %d of %lf\n", r, log_r, next_r, k, r_steps);

				if (wedge > 0) {
					ir = (pow(next_r,3.0) - pow(r, 3.0))/3.0;
					id = cos((90 - nu - nu_step)/deg) - cos((90 - nu)/deg);

					V = ir * id * mu_step / deg;

					ra = 0.0;
					dec = 0.0;
					point0 = 0.0;
					point1 = 0.0;

					atGCToEq(mu + 0.5 * mu_step, nu + 0.5 * nu_step, &ra, &dec, get_node(), wedge_incl(wedge));
					atEqToGal(ra, dec, &point0, &point1);
					integral_point[0] = point0;
					integral_point[1] = point1;
					integral_point[2] = (next_r+r)/2.0;

					total_volume += V;
				} else {
					V = mu_step * nu_step * r_step;
					xyz[0] = mu + 0.5*mu_step;
					xyz[1] = nu + 0.5*nu_step;
					xyz[2] = r + 0.5*r_step;
					xyz2lbr(xyz, integral_point);
				}

				double bg_prob = 0.0;
				double st_prob = 0.0;

				if (convolve) {
//					bg_prob = Probability.stPbxConvolved(integral_point, background_parameters);
//					st_prob = Probability.stPsgConvolved(integral_point, stream_parameters[0], wedge);
				} else {
					bg_prob = stPbx(integral_point, background_parameters);
					st_prob = stPsg(integral_point, stream_parameters[0]);
				}

				background_integral += bg_prob * V;
				stream_integral += st_prob * V;
				num_calls++;

//				if (num_calls%1000000 == 0) {
//					fprintf(stderr, "integral point: %lf %lf %lf\n", integral_point[0], integral_point[1], integral_point[2]);
//					fprintf(stderr, "ir: %lf, id: %lf, deg: %lf\n", ir, id, deg);
//					fprintf(stderr, "V: %lf, bg_prob: %lf, st_prob: %lf\n", V, bg_prob, st_prob);
//					fprintf(stderr, "bg_prob: %lf, st_prob: %lf, total_volume: %lf\n", bg_prob, st_prob, total_volume);
//					fprintf(stderr, "completed %d calls. background_integral: %lf, stream_integral: %lf\n", num_calls, background_integral, stream_integral);
//				}
			}
		}
	}
	stream_integrals[0] = stream_integral;

	finish = MPI_Wtime();
	
//	fprintf(stderr, "Integral Function Completed in %lf, Number Calls: %d, background_integral: %f, stream_integral: %f\n", (finish-start), num_calls, background_integral, stream_integral);
	integral_time = finish-start;
}
