#include <time.h>
#include <stdlib.h>

/********
	*	GML includes
*********/
#include	"gml.h"
#include	"evaluator.h"

/********
	*	MPI includes
 ********/
#include	"mpi.h"

/********
	*	Global Variables
 ********/
int		integral_defined = 0;
double*		(*integral_function)(double*) = NULL;
double*		(*integral_combinator)(double*, int) = NULL;
int		integral_parameter_length;
int		integral_results_length;

double*		(*likelihood_function)(double*) = NULL;
double		(*likelihood_combinator)(double*, int) = NULL;
int		likelihood_parameter_length;
int		likelihood_results_length;

char    	hostname[MPI_MAX_PROCESSOR_NAME];
int     	hostname_length;
int     	rank, number_processes;

double*		integral_parameters;
double*		likelihood_parameters;
double*		integral_results_recv;
double*		integral_results_send;
double*		likelihood_results_recv;
double*		likelihood_results_send;

FILE*		evaluation_file;
long		evaluations_done = 0;
int		max_evaluations = 100000;
double		best_likelihood = 100000;

void reset_evaluator() {
	evaluations_done = 0;
	best_likelihood = 100000;
}
                
void evaluator__max_evaluations(int m_e) {
	max_evaluations = m_e;
}

void set_evaluation_output(char* name) {
        evaluation_file = fopen(name, "w+");
        if (evaluation_file < 0) {
                fprintf(stderr, "Unable to open evaluation output file %s\n", name);
        }
}

double track_evaluate(double* parameters) {
        double start_time = MPI_Wtime();
        double result = evaluate(parameters);
        
        if (result < best_likelihood) {
                best_likelihood = result;
        }
        printf("%ld : %lf : %lf : %lf\n", evaluations_done, best_likelihood, result, (MPI_Wtime()-start_time));
        evaluations_done++;
 
        if (evaluation_file != NULL && evaluation_file > 0) {
                fprintf(evaluation_file, "%ld : %lf : %lf\n", evaluations_done, result, (MPI_Wtime() - start_time));  
        }
         
        return result;
}

void evaluator__init(int number_arguments, char** arguments) {
        MPI_Init(&number_arguments, &arguments);
        
        MPI_Comm_rank(MPI_COMM_WORLD, &rank);
        MPI_Comm_size(MPI_COMM_WORLD, &number_processes);
        MPI_Get_processor_name(hostname, &hostname_length);
}


void evaluator__init_data(void (*read_data)(int, int)) {
	read_data(rank, number_processes);
}

void evaluator__init_integral(double* (*i_f)(double*), int i_p_l, double* (*i_c)(double*, int), int i_r_l) {
	integral_defined = 1;

	integral_function = (*i_f);
	integral_combinator = (*i_c);
	integral_parameter_length = i_p_l;
	integral_results_length = i_r_l;

	integral_parameters = (double*)malloc(sizeof(double) * integral_parameter_length);
	integral_results_send = (double*)malloc(sizeof(double) * integral_results_length);
	integral_results_recv = (double*)malloc(sizeof(double) * (number_processes) * integral_results_length);
}

void evaluator__init_likelihood(double* (*l_f)(double*), int l_p_l, double (*l_c)(double*, int), int l_r_l) {
	likelihood_function = (*l_f);
	likelihood_combinator = (*l_c);
	likelihood_parameter_length = l_p_l;
	likelihood_results_length = l_r_l;

	likelihood_parameters = (double*)malloc(sizeof(double) * likelihood_parameter_length);
	likelihood_results_send = (double*)malloc(sizeof(double) * likelihood_results_length);
	likelihood_results_recv = (double*)malloc(sizeof(double) * (number_processes));

	int evaluations = 0;

	if (rank == 0) {
		return;
	} else {
		while (evaluations < max_evaluations) {
			if (integral_defined > 0) {
				MPI_Bcast(integral_parameters, integral_parameter_length, MPI_DOUBLE, 0, MPI_COMM_WORLD);
				integral_results_send = (*integral_function)(integral_parameters);
				MPI_Gather(integral_results_send, integral_results_length, MPI_DOUBLE, integral_results_recv, integral_results_length, MPI_DOUBLE, 0, MPI_COMM_WORLD);
			}
			MPI_Bcast(likelihood_parameters, likelihood_parameter_length, MPI_DOUBLE, 0, MPI_COMM_WORLD);
			likelihood_results_send = (*likelihood_function)(likelihood_parameters);
			MPI_Gather(likelihood_results_send, likelihood_results_length, MPI_DOUBLE, likelihood_results_recv, likelihood_results_length, MPI_DOUBLE, 0, MPI_COMM_WORLD);
			evaluations++;
		}
		exit(0);
	}
}

double evaluate(double* parameters) {
	double start, finish;
	double result;
	int i;

	if (rank == 0) {
		if (integral_defined > 0) {
			integral_parameters = parameters;
			MPI_Bcast(integral_parameters, integral_parameter_length, MPI_DOUBLE, 0, MPI_COMM_WORLD);
			integral_results_send = (*integral_function)(integral_parameters);
			MPI_Gather(integral_results_send, integral_results_length, MPI_DOUBLE, integral_results_recv, integral_results_length, MPI_DOUBLE, 0, MPI_COMM_WORLD);
			likelihood_parameters = (*integral_combinator)(integral_results_recv, number_processes);
		} else {
			likelihood_parameters = parameters;
		}
		MPI_Bcast(likelihood_parameters, likelihood_parameter_length, MPI_DOUBLE, 0, MPI_COMM_WORLD);
		likelihood_results_send = (*likelihood_function)(likelihood_parameters);
		MPI_Gather(likelihood_results_send, likelihood_results_length, MPI_DOUBLE, likelihood_results_recv, likelihood_results_length, MPI_DOUBLE, 0, MPI_COMM_WORLD);
		result = (*likelihood_combinator)(likelihood_results_recv, number_processes);

		return result;
	} else {
		fprintf(gml_get_output(), "ERROR: calling evaluate from non-master process.");
		return -1.0;
	}
}
