#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include <string.h>

#include "mpi.h"

#include "gml.h"
#include "evaluator.h"
#include "genetic_search.h"
#include "matrix.h"

int	rank;
int	number_processes;
char	hostname[MPI_MAX_PROCESSOR_NAME];
int	hostname_length;

typedef struct parameters {
	double likelihood;
	double* params;
	struct parameters *next;
	struct parameters *prev;
} PARAMETERS;

int	max_iterations = 1000;
int	number_mutations = 10;
int	population_size = 100;
int	elite_population_size = 10;

int	parameter_length;
double*	parameter_min;
double*	parameter_max;

void genetic_search__max_iterations(int mi) {
	max_iterations = mi;
}

double* get_random_parameters() {
	int i;
	double* new_parameters = (double*)malloc(sizeof(double) * parameter_length);

	for (i = 0; i < parameter_length; i++) {
		new_parameters[i] = (drand48() * (parameter_max[i]-parameter_min[i])) + parameter_min[i];
	}
	return new_parameters;
}

PARAMETERS copy(PARAMETERS p) {
	int i;
	PARAMETERS new_parameters;
	new_parameters.params = (double*)malloc(sizeof(double) * parameter_length);
	for (i = 0; i < parameter_length; i++) {
		new_parameters.params[i] = p.params[i];
	}
	new_parameters.likelihood = p.likelihood;
	new_parameters.next = NULL;
	new_parameters.prev = NULL;
	return new_parameters;
}

int equals(PARAMETERS p1, PARAMETERS p2) {
	int i;
	for (i = 0; i < parameter_length; i++) {
		if (p1.params[i] != p2.params[i]) return 0;
	}
	return 1;	
}

PARAMETERS reproduce(PARAMETERS p1, PARAMETERS p2) {
	/********
		*	Average parameters
	 ********/
	int i;
	PARAMETERS new_parameters = copy(p1);
	for (i = 0; i < parameter_length; i++) {
		new_parameters.params[i] += p2.params[i];
		new_parameters.params[i] /= 2.0;
	}
	return new_parameters;
}

PARAMETERS mutate(PARAMETERS p) {
	int target = drand48() * parameter_length;
	PARAMETERS new_parameters = copy(p);
	new_parameters.params[target] = (drand48() * (parameter_max[target]-parameter_min[target])) + parameter_min[target];
	return new_parameters;
}

void merge_sort(PARAMETERS* parameters, int first, int last) {
	int i, first_iterator, second_iterator, middle;
	PARAMETERS* sorted;

	if (last-first > 1) {
		middle = (last+first)/2;
		merge_sort(parameters, first, middle);
		merge_sort(parameters, middle, last);

		sorted = (PARAMETERS*)malloc(sizeof(PARAMETERS) * (last-first));
		first_iterator = first;
		second_iterator = middle;
		for (i = 0; i < (last-first); i++) {
			if (first_iterator >= middle) {
				sorted[i] = parameters[second_iterator];
				second_iterator++;
			} else if (second_iterator >= last) {
				sorted[i] = parameters[first_iterator];
				first_iterator++;
			} else if (parameters[second_iterator].likelihood < parameters[first_iterator].likelihood) {
				sorted[i] = parameters[second_iterator];
				second_iterator++;
			} else {
				sorted[i] = parameters[first_iterator];
				first_iterator++;
			}
		}
		for (i = 0; i < (last-first); i++) {
			parameters[first+i] = copy(sorted[i]);
			free(sorted[i].params);
		}
		free(sorted);
	}
}

double* genetic_search__iterative(double* p_min, double* p_max, int p_length) {
	int i, current_prev, current_next, current_iteration;
	PARAMETERS* population;
	PARAMETERS* next_population;
	PARAMETERS* temp_population;

	reset_evaluator();

	parameter_length = p_length;
	parameter_min = p_min;
	parameter_max = p_max;

	population = (PARAMETERS*)malloc(sizeof(PARAMETERS) * population_size);
	next_population = (PARAMETERS*)malloc(sizeof(PARAMETERS) * population_size);
	temp_population = (PARAMETERS*)malloc(sizeof(PARAMETERS) * population_size);

	for (i = 0; i < population_size; i++) {
		population[i].params = get_random_parameters();
		population[i].likelihood = track_evaluate(population[i].params);
	}

	merge_sort(population, 0, population_size);
	for (current_iteration = 0; current_iteration < max_iterations; current_iteration++) {
		printf("iteration: %d\n", current_iteration);
		/********
			*	Generate the next population
		 ********/
		i = 0;
		while (i < number_mutations) {
			next_population[i] = mutate(population[i]);
			next_population[i].likelihood = track_evaluate(next_population[i].params);
			i++;
		}
		int first = 0;
		while (i < population_size) {
			int second = population_size * drand48();
			while (second == first) second = population_size * drand48();
		
			next_population[i] = reproduce(population[first], population[second]);
			next_population[i].likelihood = track_evaluate(next_population[i].params);

			first++;
			if (first >= elite_population_size) first = 0;		
			i++;
		}

		merge_sort(next_population, 0, population_size);

		current_prev = 0;
		current_next = 0;
		for (i = 0; i < population_size; i++) {
			if (current_next >= population_size) {
				temp_population[i] = population[current_prev];
				current_prev++;
			} else if (current_prev >= population_size) {
				temp_population[i] = population[current_next];
				current_next++;
			} else if (population[current_prev].likelihood < population[current_next].likelihood) {
				temp_population[i] = population[current_prev];
				current_prev++;
			} else if (population[current_next].likelihood < population[current_prev].likelihood) {
				temp_population[i] = population[current_next];
				current_next++;
			} else if (equals(population[current_prev], population[current_next])) {
				temp_population[i] = population[current_next];
				current_next++;
				current_prev++;
			} else {
				temp_population[i] = population[current_next];
				current_next++;
			}

//			printf("%d: %lf\n", i, temp_population[i].likelihood);
		}
		population = temp_population;
	}
}

/********
	*	Probabilistic Genetic Search
 ********/


int current_population_size;
int max_population_size;
double **population;
double *likelihoods;
double reproduction_chance;

void get_next_params(double*** params, int *number) {
	if (current_population_size < max_population_size) {
		*number = 1;
		*params = (double**)malloc(sizeof(double*));
		(*params)[0] = get_random_parameters();
		return;
	}

	double generate_reproduction = drand48();
	if (generate_reproduction < reproduction_chance) {
		int i, j, k;
		int first = (int)(drand48() * current_population_size);
		int second = (int)(drand48() * current_population_size);
		while (second == first) {
			second = (int)(drand48() * current_population_size);
		}

		*number = 1;
		*params = (double**)malloc(sizeof(double*) * 1);

		(*params)[0] = (double*)malloc(sizeof(double) * parameter_length);	//average
//		(*params)[1] = (double*)malloc(sizeof(double) * parameter_length);	//prev
//		(*params)[2] = (double*)malloc(sizeof(double) * parameter_length);	//next

//		fprintf(stderr, "parameters generated:\n");
		for (i = 0; i < parameter_length; i++) {
			(*params)[0][i] = (population[first][i] + population[second][i]) / 2.0;
		//	double diff = (*params)[0][i] - population[first][i];
		//	(*params)[1][i] = population[first][i] - diff;		//subtract if first < params, add otherwise
		//	(*params)[2][i] = population[second][i] + diff;		//add if first < params, subtract otherwise

//			fprintf(stderr, "\tavg: %lf, min %lf, max: %lf\n", (*params)[0][i], (*params)[1][i], (*params)[2][i]);
		}
	} else {
		*number = 1;
		*params = (double**)malloc(sizeof(double*));
		int target = (int)(drand48() * current_population_size);

		(*params)[0] = (double*)malloc(sizeof(double) * parameter_length);
		memcpy((*params)[0], population[target], sizeof(double) * parameter_length);

		int target_param = (int)(drand48() * parameter_length);
		(*params)[0][target_param] = parameter_min[target_param] + ((parameter_max[target_param]-parameter_min[target_param]) * drand48());
	}
}

double insert_params(double* params) {
	double result = evaluate(params);
	if (isnan(result)) return result;

	if (current_population_size == 0) {
		population = (double**)malloc(sizeof(double*));
		population[0] = (double*)malloc(sizeof(double) * parameter_length);
		memcpy(population[0], params, sizeof(double) * parameter_length);

		likelihoods = (double*)malloc(sizeof(double));
		likelihoods[0] = result;

		current_population_size = 1;
	} else {
		int position = current_population_size;
		int i;

		for (i = 0; i < current_population_size; i++) {
			if (result < likelihoods[i]) {
				position = i;
				break;
			} else if (result == likelihoods[i]) {
				return result;
			}
		}

		if (current_population_size < max_population_size) {
			current_population_size++;
			population = (double**)realloc(population, sizeof(double*) * current_population_size);
			likelihoods = (double*)realloc(likelihoods, sizeof(double) * current_population_size);

			population[current_population_size-1] = (double*)malloc(sizeof(double) * parameter_length);
		}

		if (position < current_population_size) {
			for (i = current_population_size-1; i > position; i--) {
				memcpy(population[i], population[i-1], sizeof(double) * parameter_length);
				likelihoods[i] = likelihoods[i-1];
			}
			memcpy(population[position], params, sizeof(double) * parameter_length);
			likelihoods[position] = result;
		}
	}
	return result;
}

double* genetic_search__probabilistic(double* p_min, double* p_max, int p_length, int m_p_s, double r_c, int max_evaluations) {
	int evaluations_done = 0;
	parameter_min = p_min;
	parameter_max = p_max;
	parameter_length = p_length;
	reproduction_chance = r_c;
	max_population_size = m_p_s;
	current_population_size = 0;

	while (evaluations_done < max_evaluations) {
		double **params;
		int number_params;
		int i, j, k;
		get_next_params(&params, &number_params);

		for (i = 0; i < number_params; i++) {
			double result = insert_params(params[i]);
			evaluations_done++;

/*			for (j = 0; j < current_population_size; j++) {
				fprintf(stderr, "\t[%ld]: %lf :", j, likelihoods[j]);

				for (k = 0; k < parameter_length; k++) {
					fprintf(stderr, " %lf", population[j][k]);
				}
				fprintf(stderr, "\n");
			}
*/
			fprintf(stderr, "%ld: min: %lf, max: %lf, median: %lf, current: %lf\n", evaluations_done, likelihoods[0], likelihoods[current_population_size-1], likelihoods[current_population_size/2], result);
		}

		for (i = 0; i < number_params; i++) {
			free(params[i]);
		}
		free(params);
	}
	return population[0];
}
