#include <stdio.h>
#include <time.h>
#include <math.h>
#include <string.h>

#include "mpi.h"

#include "gml.h"
#include "evaluator.h"
#include "gradient_descent.h"
#include "matrix.h"

#define	CONJUGATE_GRADIENT_DESCENT 1
#define GRADIENT_DESCENT 0
#define CONJUGATE_GRADIENT_HESSIAN_DESCENT 0

/********
	*	Stopping conditions
 ********/
double		line_search_x_cutoff = 0.00000001;
double		line_search_fx_cutoff = 0.00000001;
double		tol = 0.000001;
double		EPS = 0.00000003;
double		gradient_cutoff = 0.0002;
int		max_line_searches = 10000;
int		brent_iterations = 1000;
int		line_search_iterations = 10000;
double		minimum_probability = 0.0001;
double		minimum_probability_change = 0.00000001;
double*		previous_parameters;
double		previous_probability;
double*		current_parameters;
double		current_probability;

double*		current_gradient;
double*		direction;
double*		previous_direction;

int	rank;
int	number_processes;
char	hostname[MPI_MAX_PROCESSOR_NAME];
int	hostname_length;
int	descent_type = CONJUGATE_GRADIENT_DESCENT;
long	gradients_calculated = 0;

double	start_time;
int	quit_cgd = 0;

double get_gradient(double *parameters, int parameter, double step_size) {
	double v1 = 0.0;
	double v2 = 0.0;

	parameters[parameter] += step_size;
	v1 = track_evaluate(parameters);
	parameters[parameter] -= step_size*2;
	v2 = track_evaluate(parameters);
	parameters[parameter] += step_size;

	return (v1-v2) / (2.0 * step_size);
}

MATRIX malik_line_search(MATRIX x0, MATRIX d_ser, double tols, int nquad, double step) {
	MATRIX d = matrix_apply_div(d_ser, matrix_norm(d_ser));
	int jump = 2;
	int m = 0;
	double e1, e2, e3;
	MATRIX x1, x2, x3;
	int i;

	e1 = track_evaluate(matrix2array(x0));
	while (m < 1) {
		MATRIX x2 = matrix_add(x0, matrix_apply_mul(d, step));
		e2 = track_evaluate(matrix2array(x2));

		if (e2 < e1) {
			m = 2;
		} else {
			if (step < tol) {
				fprintf(stderr, "line search could not progress beyond tolsearch\n");
				quit_cgd = 1;
				return;
			}
			step = step / 2;
		}
	}

	double l1 = 0;
	double l2 = 1;
	double l3 = 2;
	while (m < 3) {
//		x1 = x0 + l1*step*d;
		x1 = matrix_add(x0, matrix_apply_mul(d, l1*step));
		e1 = track_evaluate(matrix2array(x1));

//		x2 = x0 + l2 * step * d;
		x2 = matrix_add(x0, matrix_apply_mul(d, l2*step));
		e2 = track_evaluate(matrix2array(x2));

//		x3 = x0 + l3 * step * d;
		x3 = matrix_add(x0, matrix_apply_mul(d, l3*step));
		e3 = track_evaluate(matrix2array(x3));

//		fprintf(stderr, "e3: %lf, e2: %lf\n", e3, e2);

		if (e3 >= e2) {
//			fprintf(stderr, "2nd while completed.\n");
			m = 4;
		} else if (isnan(e3-e3)) {
			fprintf(stderr, "premature exit of linesearch due to NaN\n");
			quit_cgd = 1;
			return;
		} else {
			l1 = l2;
			l2 = l3;
			l3 = jump * l3;
		}
	}

	MATRIX x;
	int chtst = 0;
	double es;
	for (i = 1; i < nquad; i++) {
		if (chtst < 5) {
			double a = (l1*l1)*(e2-e3) + (l2*l2)*(e3-e1) + (l3*l3)*(e1-e2);
			double b = l1*(e2-e3) + l2*(e3-e1) + l3*(e1-e2);
			double lstar = 0.5*(a/b);

//			xs = x0 + lstar*step*d;
			MATRIX xs = matrix_add(x0, matrix_apply_mul(d, (lstar*step)));
			es = track_evaluate(matrix2array(xs));

//			fprintf(stderr, "es: %lf, lstar: %lf, l1: %lf, l2: %lf, l3: %lf\n", es, lstar, l1, l2, l3);

			if (isnan(es-es)) {
				fprintf(stderr, "premature exit of linesearch due to NaN 2\n");
				quit_cgd = 1;
				return;
			}
			if (chtst < 5) {
				if (lstar > l2 + tol) {
					if (es > e2) {
						l3 = lstar;
						e3 = es;
					} else {
						l1 = l2;
						e1 = e2;
						l2 = lstar;
						e2 = es;
					}
				} else if (lstar < l2 - tol) {
					if (es > e2) {
						l1 = lstar;
						e1 = es;
					} else {
						l3 = l2;
						e3 = e2;
						l2 = lstar;
						e2 = es;
					}
				} else {
					lstar = l2 + tol;
//					es = evaluate(x0 + lstar*step*d);
					es = track_evaluate(matrix2array( matrix_add(x0, matrix_apply_mul(d, (lstar*step))) ));

//					fprintf(stderr, "case 3: es: %lf, e2: %lf\n", es, e2);
					if (es < e2) {
						l1 = l2;
						e1 = e2;
						l2 = lstar;
						e2 = es;
					} else {
						l3 = lstar;
						e3 = es;
					}
				}
//				x = x0 + lstar*step*d;
				x = matrix_add(x0, matrix_apply_mul(d, (lstar*step)));
			}
		}
	}
	printf("finished line search at probability: %lf\n", es);
	return x;
}

double* malik_cgd(double* step_sizes, double* initial_parameters, int parameter_length) {
	MATRIX h, h_prev, g_prev, d_prev, d_pres, x_pres, d_elg, d_elx, g_pres;
	int i, j;

	h_prev = create_matrix(parameter_length, parameter_length);
	g_prev = create_matrix(parameter_length, 1);
	d_prev = create_matrix(parameter_length, 1);

	for (i = 0; i < parameter_length; i++) {
		h_prev.values[i][i] = 1.0;
	}

	x_pres = create_matrix(parameter_length, 1);
	for (i = 0; i < parameter_length; i++) {
		x_pres.values[i][0] = initial_parameters[i];
	}
	d_elg = create_matrix(parameter_length, 1);
	d_elx = create_matrix(parameter_length, 1);

	int nquad = 3;
	double f_fut, d_elf, f_pres;
	double targ = 0.1;
	double tolgrad = 0.00000001;
	double tolsearch = 0.000000001;
	double tol = tolgrad;
	double tols = tolsearch;

	f_pres = track_evaluate(initial_parameters);

	if (isnan(f_pres)) {
		printf("0 : NAN");
		return;
	}

	for (i = 0; i < max_line_searches; i++) {
		g_pres = create_matrix(parameter_length, 1);
		double max_gradient = 0;
//		printf("gradient:\n");
		for (j = 0; j < parameter_length; j++) {
			g_pres.values[j][0] = get_gradient(matrix2array(x_pres), j, step_sizes[j]);
			if (abs(g_pres.values[j][0]) > max_gradient) max_gradient = abs(g_pres.values[j][0]);
//			printf("\t[%d]: %lf\n", j, g_pres.values[j][0]);
		}
//		if (max_gradient < line_search_x_cutoff) {
//			fprintf(stderr, "max_gradient < line_search_x_cutoff -- %lf < %lf\n", max_gradient, line_search_x_cutoff);
//			return;
//		}

		d_elg = matrix_sub(g_pres, g_prev);

		if (CONJUGATE_GRADIENT_HESSIAN_DESCENT) {
			//t1 = h_prev * d_elg;
			MATRIX t1 = matrix_mul(h_prev, d_elg);

			//t2 = d_elg' * t1;
			MATRIX t2 = matrix_mul(invert(d_elg), t1);

			//t3 = d_elx' * d_elg;
			MATRIX t3 = matrix_mul(invert(d_elx), d_elg);

			//u = d_elx/t3 - t1/t2;
			MATRIX temp1 = matrix_rdiv(d_elx, t3);
			MATRIX temp2 = matrix_rdiv(t1, t2);
			MATRIX u = matrix_sub(temp1, temp2);

			//h = h_prev + (d_elx*d_elx')/t3 - (t1*t1')/t2 + t2*(u*u');
			MATRIX tt = matrix_mul(d_elx, invert(d_elx));
			MATRIX h_1 = matrix_rdiv(tt, t3);
			MATRIX h_2 = matrix_rdiv(matrix_mul(t1, invert(t1)), t2);
			MATRIX h_3 = matrix_mul(matrix_mul(u, invert(u)), t2);
			h = matrix_add(matrix_sub(matrix_add(h_prev, h_1), h_2), h_3);

			double hm = mean(h);
			if (isnan(hm)) {
				fprintf(stderr, "HM is NAN, quitting\n");
				return;
			}
			h_prev = h;
		}

		int bold = 10;
		int nhessian = 10000;
		int nreset = 10;
		double ng, step;
		double lam = 0.1;
		double tolf = tol * tols;
		if (CONJUGATE_GRADIENT_DESCENT) {
			MATRIX bet;
			if (matrix_norm(g_prev) == 0 || (i%nreset) == 0) {
				bet = create_matrix(1,1);
			} else {
				//bet = g_pres'*(g_pres-g_prev)/(g_prev'*g_prev);
				bet = matrix_rdiv(matrix_mul(invert(g_pres), matrix_sub(g_pres, g_prev)), matrix_mul(invert(g_prev), g_prev));
			}

			//d_pres = -g_pres + bet*d_prev;
			d_pres = matrix_add(matrix_neg(g_pres), matrix_mul(bet, d_prev));

			ng = matrix_norm(g_pres);
//			step = fabs(f_pres - targ)/ng;
			step = 0.5;
		} else if (GRADIENT_DESCENT) {
			d_pres = matrix_neg(g_pres);
			step = 0.5;
			ng = matrix_norm(g_pres);
		} else if (CONJUGATE_GRADIENT_HESSIAN_DESCENT) {
			//bet=gpres'*delg/(gprev'*gprev);
			MATRIX temp1 = matrix_mul(invert(g_pres), d_elg);
			MATRIX temp2 = matrix_mul(invert(g_prev), g_prev);
			MATRIX bet = matrix_rdiv(temp1, temp2);
			if (isinf(bet.values[0][0])) bet.values[0][0] = 0;

			//d1=-gpres+bet*dprev;
			//nd1=norm(d1);
			MATRIX d1 = matrix_add(matrix_neg(g_pres), matrix_mul(bet, d_prev));
			double nd1 = matrix_norm(d1);

			//d2=-H*gpres;
			//nd2=norm(d2);
			MATRIX d2 = matrix_mul(matrix_neg(h), g_pres);
			double nd2 = matrix_norm(d2);

			//dpres=(lam*d1/nd1+d2/nd2)/(lam+1);
			MATRIX d_temp1 = matrix_apply_mul(d1, lam/nd1);
			MATRIX d_temp2 = matrix_apply_div(d2, nd2);
			d_pres = matrix_apply_div(matrix_add(d_temp1, d_temp2),(lam+1));

			ng = matrix_norm(g_pres);
			step = (lam*abs(f_pres-targ)/ng+nd2)/(lam+1);
		}

		MATRIX x_fut;
		if (ng > tol) {
//			fprintf(stderr, "iteration %d, f_pres: %lf, ng: %lf, targ: %lf, step: %lf, ", i, f_pres, ng, targ, step);
			printf("Starting line search: %d\n",i);
			x_fut = malik_line_search(x_pres, d_pres, tols, nquad, step);
			if (quit_cgd > 0) {
				fprintf(stderr, "could not complete line search, quitting.\n");
				return matrix2array(x_pres);
			}
		} else if (isnan(ng-ng)) {
			x_fut = x_pres;
			fprintf(stderr, "ng is NaN, quitting\n");
			return matrix2array(x_pres);
		} else {
			x_fut = x_pres;
			fprintf(stderr, "gradient is below tolgrad, quitting\n");
			return matrix2array(x_pres);
		}

		f_fut = track_evaluate(matrix2array(x_fut));

//		if (abs(f_fut-f_pres) < line_search_fx_cutoff) {
//			fprintf(stderr, "abs(f_fut - f_pres) < line_search_fx_cutoff -- %lf < %lf\n", abs(f_fut-f_pres), line_search_fx_cutoff);
//			return;
//		}

		d_elf = f_fut - f_pres;
		if (CONJUGATE_GRADIENT_HESSIAN_DESCENT) {
			if (d_elf >= -tolf) {
				lam = bold^10;
				nhessian = i+5;
			} else {
				if (lam > bold^(-10)) {
        		                lam = lam/bold;
				}
			}
		}
		d_elx = matrix_sub(x_fut, x_pres);
		x_pres = x_fut;
		f_pres = f_fut;
		d_prev = d_pres;
		g_prev = g_pres;
	}
	return matrix2array(x_pres);
}

double* conjugate_gradient_descent(double* initial_parameters, double* step_sizes, int parameter_length) {
	int i;
	printf("NEW CONJUGATE_GRADIENT_DESCENT\nparameters:");
	for (i = 0; i < parameter_length; i++) {
		printf(" %lf", initial_parameters[i]);
	}
	printf("\n");
	start_time = MPI_Wtime();
	quit_cgd = 0;

	reset_evaluator();
	return malik_cgd(step_sizes, initial_parameters, parameter_length);
}
