package gmle.salsa.astronomy;

public class ATSurveyGeometry {
	static double D2PI = 6.2831853071795864769252867665590057683943387987502;
	static double DPI = 3.1415926535897932384626433832795028841971693993751;
	static double at_stripeSeperation = 2.5;
	static double at_surveyCenterRa = 185.0;
	static double at_surveyCenterDec = 32.5;
	static double at_deg2Rad = DPI/180.0;
	static double at_rad2Deg = 180/DPI;

	static double dmod(double A, double B) {
		return ((B)!=0.0?((A)*(B)>0.0?(A)-(B)*java.lang.Math.floor((A)/(B)):(A)+(B)*java.lang.Math.floor(-(A)/(B))):(A));
	}

	static double dsign(double A, double B) {
		return ((B)<0.0?-(A):(A));
	}

	/*	IN amu -- mu in degrees
		IN anu -- nu in degrees
		OUT ra -- ra in degrees
		OUT dec -- dec in degrees
		IN anode -- node in degrees
		IN ainc -- inclination in degrees
	*/
	static void atGCToEq(double amu, double anu, double[] ra, double[] dec, double anode, double ainc) {
		double x1, y1, z1, x2, y2, z2;
		/* Convert to radians */

		amu = amu * at_deg2Rad;
		anu = anu * at_deg2Rad;
		anode = anode * at_deg2Rad;
		ainc = ainc * at_deg2Rad;
		/* Rotation */
		x2 = java.lang.Math.cos(amu-anode)*java.lang.Math.cos(anu);
		y2 = java.lang.Math.sin(amu-anode)*java.lang.Math.cos(anu);
		z2 = java.lang.Math.sin(anu);
		x1 = x2;
		y1 = y2*java.lang.Math.cos(ainc) - z2*java.lang.Math.sin(ainc);
		z1 = y2*java.lang.Math.sin(ainc) + z2*java.lang.Math.cos(ainc);

		ra[0] = java.lang.Math.atan2 (y1, x1)+anode;
		dec[0] = java.lang.Math.asin(z1);
		/* Convert back to degrees */
		ra[0] = ra[0] * at_rad2Deg;
		dec[0] = dec[0] * at_rad2Deg;

		ATSurveyGeometry.atBound2(dec, ra);
		return;
	}

	/*
		double ra,	 IN -- ra in degrees
		double dec,	 IN -- dec in degrees 
		double *glong,	 OUT -- Galactic longitude in degrees
		double *glat	 OUT -- Galactic latitude in degrees
	*/
	static void atEqToGal (double ra, double dec, double[] glong, double[] glat) {
		/* Convert to radians */
		ra = ra * at_deg2Rad;
		dec = dec * at_deg2Rad;
		/* Use SLALIB to do the actual conversion */
		slaEqgal(ra, dec, glong, glat);
		/* Convert back to degrees */
		glong[0] = glong[0] * at_rad2Deg;
		glat[0] = glat[0] * at_rad2Deg;
		ATSurveyGeometry.atBound2(glat, glong);
		return;
	}

	/*
			  double *angle,	MODIFIED -- the angle to bound in degrees
			  double min,	IN -- inclusive minimum value 
			  double max	IN -- exclusive maximum value 
	*/
	static void atBound(double[] angle, double min, double max) {
		while (angle[0] < min) {
			angle[0] += 360.0;
		}
		while (angle[0] >= max) {
			angle[0] -= 360.0;
		}
		return;
	}

	/*
	      double *theta,	MODIFIED -- the -90 to 90 angle
	      double *phi	MODIFIED -- the 0 to 360 angle
	*/
	static void atBound2(double[] theta, double[] phi) {
		atBound(theta, -180.0, 180.0);
		if (java.lang.Math.abs(theta[0]) > 90.0) {
			theta[0] = 180.0 - theta[0];
			phi[0] += 180;
		}
		atBound(theta, -180.0, 180.0);
		atBound(phi, 0.0, 360.0);
		if (java.lang.Math.abs(theta[0])==90.0) phi[0]=0.;
		return;
	}

	static void slaDcc2s(double[] v, double[] a, double[] b) {
		double x, y, z, r;

		x = v[0];
		y = v[1];
		z = v[2];
		r = java.lang.Math.sqrt ( x * x + y * y );

		a[0] = ( r != 0.0 ) ? java.lang.Math.atan2 ( y, x ) : 0.0;
		b[0] = ( z != 0.0 ) ? java.lang.Math.atan2 ( z, r ) : 0.0;
	}


	static void slaDcs2c(double a, double b, double[] v) {
		double cosb;

		cosb = java.lang.Math.cos(b);
		v[0] = java.lang.Math.cos(a) * cosb;
		v[1] = java.lang.Math.sin(a) * cosb;
		v[2] = java.lang.Math.sin(b);
	}


	static void slaDimxv(double[][] dm, double[] va, double vb[]) {
		double w;
		double[] vw = new double[3];

		/* Inverse of matrix dm * vector va -> vector vw */
		for (int j = 0; j < 3; j++) {
			w = 0.0;
			for (int i = 0; i < 3; i++) {
				w += dm[i][j] * va[i];
			}
			vw[j] = w;
		}

		/* Vector vw -> vector vb */
		for (int j = 0; j < 3; j++) {
			vb[j] = vw[j];
		}
	}


	static void slaDmxv(double[][] dm, double va[], double[] vb) {
		double w;
		double[] vw = new double[3];

		/* Matrix dm * vector va -> vector vw */
		for (int j = 0; j < 3; j++) {
			w = 0.0;
			for (int i = 0; i < 3; i++) {
				w += dm[j][i] * va[i];
			}
			vw[j] = w;
		}

		/* Vector vw -> vector vb */
		for (int j = 0; j < 3; j++) {
			vb[j] = vw[j];
		}
	}


	static double slaDrange(double angle) {
		double w;

		w = ATSurveyGeometry.dmod ( angle, D2PI );
		return ( java.lang.Math.abs ( w ) < DPI ) ? w : w - ATSurveyGeometry.dsign ( D2PI, angle );
	}


	static double slaDranrm(double angle) {
		double w;

		w = ATSurveyGeometry.dmod ( angle, D2PI );
		return ( w >= 0.0 ) ? w : w + D2PI;
	}


	static void slaEqgal(double dr, double dd, double[] dl, double[] db) {
		double[] v1 = new double[3];
		double[] v2 = new double[3];

		double[][] rmat = new double[3][3];

		rmat[0][0] = -0.054875539726;
		rmat[0][1] = -0.873437108010;
		rmat[0][2] = -0.483834985808;
		rmat[1][0] =  0.494109453312;
		rmat[1][1] = -0.444829589425;
		rmat[1][2] =  0.746982251810;
		rmat[2][0] = -0.867666135858;
		rmat[2][1] = -0.198076386122;
		rmat[2][2] =  0.455983795705;

		/* Spherical to Cartesian */
		ATSurveyGeometry.slaDcs2c ( dr, dd, v1 );

		/* Equatorial to Galactic */
		ATSurveyGeometry.slaDmxv ( rmat, v1, v2 );

		/* Cartesian to spherical */
		ATSurveyGeometry.slaDcc2s ( v2, dl, db );

		/* Express in conventional ranges */
		dl[0] = ATSurveyGeometry.slaDranrm ( dl[0] );
		db[0] = ATSurveyGeometry.slaDrange ( db[0] );
	}
	
	/*return eta from stripe number*/
	static double atEtaFromStripeNumber(int wedge) {
		double at_stripeSeparation = 2.5;
		double eta; 
		
		if(wedge <=46) {
			eta = wedge * at_stripeSeparation - 57.5;
		} else {
			eta = wedge * at_stripeSeparation -57.5 - 180.0;
		}
			
		return eta;
	}
	
	/*return ra & dec from survey longitude and latitude*/
	static void atSurveyToEq(double slong, double slat, double[] ra, double[] dec) {
		/*convert to radians*/
		slong = slong * at_deg2Rad;
		slat = slat * at_deg2Rad;
		double anode = at_surveyCenterRa - 90.0;
		anode = anode * at_deg2Rad;
		double etaPole = at_surveyCenterDec * at_deg2Rad;
		/*Rotation*/
		double x1 = -Math.sin(slong);
		double y1 = Math.cos(slat+etaPole) * Math.cos(slong);
		double z1 = Math.sin(slat+etaPole) * Math.cos(slong);
		ra[0] = Math.atan2(y1,x1) + anode;
		dec[0] = Math.asin(z1);
		ra[0] = ra[0] * at_rad2Deg;
		dec[0] = dec[0] * at_rad2Deg;
		atBound2(dec, ra);
	}
		
}
