package gmle.salsa.astronomy;

public class Coordinates {

	private static double deg = (180.0 / java.lang.Math.PI);
	private static double[] sigmoid_curve_parameters = { 0.9402, 1.6171, 23.5877 };

	public static double deg()	{ return deg; }

	/* Convert sun-centered lbr into galactic xyz coordinates. */
	public static double[] lbr2xyz(double[] lbr) {
		double[] xyz = new double[3];

		double r0 = 8.5;
		double sinb = Math.sin( lbr[1] / deg );
		double sinl = Math.sin( lbr[0] / deg );
		double cosb = Math.cos( lbr[1] / deg );
		double cosl = Math.cos( lbr[0] / deg );

		xyz[2] = lbr[2] * sinb;
		double zp = lbr[2] * cosb;
		double d = Math.sqrt( r0 * r0 + zp * zp - 2 * r0 * zp *cosl );

		double sint = (zp * sinl) / d;
		double cost = (zp * zp - r0 * r0 - d * d) / (2 * d * r0);

		xyz[0] = d * cost;
		xyz[1] = d * sint;
		return xyz;
	}

	/* Convert galactic xyz into sun-centered lbr coordinates. */
	public static double[] xyz2lbr(double[] xyz) {
		double[] lbr = new double[3];

		double xsun = xyz[0] + 8.5;
		double temp = xsun * xsun + xyz[1] * xyz[1];

		lbr[0] = Math.atan2( xyz[1], xsun ) * deg;
		lbr[1] = Math.atan2( xyz[2], Math.sqrt( temp ) ) * deg;
		lbr[2] = Math.sqrt( temp + xyz[2] * xyz[2] );

		if (lbr[0] < 0) lbr[0] += 360;
		return lbr;
	}

	/* Get the detection efficiency at a distance of kr kiloparsecs. */
	public static double reff(double kr) {
		double gstar = r2mag(kr * 1000);
		return sigmoid_curve_parameters[0] / (Math.exp( sigmoid_curve_parameters[1] * (gstar - sigmoid_curve_parameters[2]) ) + 1);
	}

	/* Convert r in parsecs into the apparent magnitude, given an absolute magnitude of 4.2. */
	public static double r2mag(double r) {
		double absm = 4.2;
		return 5 * ((Math.log(r)/Math.log(10)) - 1) + absm;
	}
	
	/* Convert apparent magnitude, g, into r, in parsecs; given absolute magnitude absm */
	public static double mag2r(double g) {
		double absm = 4.2;
		return Math.pow(10, (((g-absm)/5.0)+1));
	}

	/* Get eta for the given wedge. */
	static double wedge_eta( int wedge ) {
		return wedge * ATSurveyGeometry.at_stripeSeperation - 57.5 - (wedge > 46 ? 180.0 : 0.0);
	}

	/* Get inclination for the given wedge. */
	static double wedge_incl( int wedge ) {
		return wedge_eta( wedge ) + ATSurveyGeometry.at_surveyCenterDec;
	}

	/* Get the node of the GC coordinates used in the survey. */
	static double get_node() {
		return ATSurveyGeometry.at_surveyCenterRa - 90.0;
	}
	
	/*Get normal vector of data slice from stripe number*/
	static void stripe_normal(int wedge, double[] xyz) {
		double[] l 	= new double[1];
		double[] b 	= new double[1];
		double[] ra 	= new double[1];
		double[] dec 	= new double[1];

		double eta = ATSurveyGeometry.atEtaFromStripeNumber(wedge);
		ATSurveyGeometry.atSurveyToEq(0, 90.0+eta, ra, dec);
		ATSurveyGeometry.atEqToGal(ra[0], dec[0], l, b);
		lbToXyz(l[0], b[0], xyz);	
	}	
	
	/*convert galactic coordinates l,b into cartesian x,y,z*/
	static void lbToXyz(double l, double b, double[] xyz) {
		l = l / deg;
		b = b / deg;
	
		xyz[0] = Math.cos(l) * Math.cos(b);
		xyz[1] = Math.sin(l) * Math.cos(b);
		xyz[2] = Math.sin(b);
	}

	/*wrapper that converts a point into magnitude-space pseudo-xyz*/	
	static double[]  xyz_mag(double[] point, double offset) {
		double[] lbg;
		lbg = xyz2lbg(point, offset);
		return lbr2xyz(lbg);
	}
	
	/*convert distance in kpcs into apparent magnatude*/
	static double[] xyz2lbg(double[]point, double offset) {
		double[] lbr;
		lbr = xyz2lbr(point);
		double g = 5.0 * ( Math.log(100.0*lbr[2]) / Math.log(10.0) ) + 4.2 - offset;
		
		lbr[2] = g;
		
		return lbr;
	}
			
}
