package gmle.salsa.astronomy;

import java.util.Random;

import flanagan.integration.IntegralFunction;
import flanagan.integration.Integration;


public class Probability {

	/*
		stPsgFunction - determines the probability that a star with given coordinates (coordpar,
		in lbr format) is a part of a stream defined by the given parameters (pars). It 
		is assumed that lbr coordinates are solar-centered and xyz coordinates are galactic 
		centered.
		Return: a double value is returned indicating the probability that the star is in 
		the stream. A higher value indicates a higher probability.
	*/
	public static double stPsgFunction(double[] star_coords, double[] stream_parameters, int wedge) {
		double[] c = new double[3];
		double[] a = new double[3];
//update: allow for new coordinate transforms
		double[] ra = new double[1];
		double[] dec = new double[1];
		double[] l = new double[1];
		double[] b = new double[1];
		double[] lbr = new double[3];
//update: load new stream parameters		
		double mu = stream_parameters[0];
		double r = stream_parameters[1];
		double theta = stream_parameters[2];
		double phi = stream_parameters[3];
		double sigma = stream_parameters[4];
//
/*
		c[0] = stream_parameters[0];
		c[1] = stream_parameters[1];
		c[2] = stream_parameters[2];
		a[0] = stream_parameters[3];
		a[1] = stream_parameters[4];
		a[2] = stream_parameters[5];
*/

//update: convert from mu, nu, r geometry to a and c geometry
                ATSurveyGeometry.atGCToEq(mu, 0, ra, dec, Coordinates.get_node(), Coordinates.wedge_incl(wedge));
                ATSurveyGeometry.atEqToGal(ra[0], dec[0], l, b);
                lbr[0] = l[0];
                lbr[1] = b[0];
                lbr[2] = r;
                c = Coordinates.lbr2xyz(lbr);

                a[0] = Math.sin(theta) * Math.cos(phi);
                a[1] = Math.sin(theta) * Math.sin(phi);
                a[2] = Math.cos(theta);
//		
//	System.err.println("vector C: " + c[0]+", "+c[1]+", "+c[2]);
//	System.err.println("vector A: " + a[0]+", "+a[1]+", "+a[2]);
		double[] xyz;

		/**
			Sigma near 0 so star prob is 0.
		**/
		if (sigma > -0.0001 && sigma < 0.0001) return 0;

		xyz = Coordinates.lbr2xyz(star_coords);

		xyz[0] = xyz[0] - c[0];
		xyz[1] = xyz[1] - c[1];
		xyz[2] = xyz[2] - c[2];

		double dotted = dot_product(a, xyz);

//		System.err.println("[STPSG] before dotted xyz: " + xyz[0] + " " + xyz[1] + " " + xyz[2]);

		xyz[0] = xyz[0] - dotted*a[0];
		xyz[1] = xyz[1] - dotted*a[1];
		xyz[2] = xyz[2] - dotted*a[2];
		

//		System.err.println("[STPSG] dotted xyz: " + xyz[0] + " " + xyz[1] + " " + xyz[2]);
		double xyz_norm = norm(xyz);

//		System.err.println("[STPSG] xyz_norm: " + xyz_norm);

		double prob = Math.exp( -(xyz_norm*xyz_norm) / 2 / (sigma*sigma) );

//		System.err.println("dotted: " + dotted + ", xyz_norm: " + xyz_norm + ", sigma: " + sigma);
//		System.err.println("[STPSG] prob before reff: " + prob);

		//double reff_value = Coordinates.reff(star_coords[2]);
//		System.err.println("reff_value: " + reff_value);
		//prob *= reff_value;

//		System.err.println("[STPSG]prob after reff: " + prob);
		return prob;
	}

	/*
		stPbxFunction - determines the probability that a star with the given coordinates
		(coordpar, in lbr format) is in the galaxy defined by the given parameters (bkgdpar)
		Return: a double that indicates the probability. Higher values indicate
		higher likelyhood that the star is a part of the galaxy distribution.
		A return < 0 indicates an error:
		-1 - q is 0
		-2 - magnitude of star is NaN
	*/
	public static double stPbxFunction(double[] star_coords, double[] background_parameters) {
		double[] xyz = Coordinates.lbr2xyz(star_coords);

		/* 
		 * alpha more negative, background falls off faster
		 * q bigger means more squashed galaxy
		 * r0, z0 defines the cylinder in the middle of the galaxy that will be empty
		 * maxz integrate up to in steps of dz, numerical consideration
		 */
		double alpha	= background_parameters[0];
		double q	= background_parameters[1];
		double r0	= background_parameters[2];
		double delta	= background_parameters[3];
   
//		System.err.println("alpha: " + alpha + ", q: " + q + ", r0: " + r0 + ", delta: " + delta);

		/* if q is 0, there is no probability */
		if (q == 0) {
			System.err.println("ERROR, Q = 0, no prbability");
			System.exit(0);
			return -1;
		}
    
		/* background probability */
		double rg = Math.sqrt( (xyz[0]*xyz[0]) + (xyz[1]*xyz[1]) + (xyz[2]*xyz[2]/(q*q)) ); 

//		System.err.println("rg: " + rg + ", xg: " + xyz[0] + " " + xyz[1] + " " + xyz[2]);

		double prob = 1.0 / (Math.pow( rg, alpha ) * Math.pow( rg + r0, 3 - alpha + delta ));

//		System.err.println("P before reff: " + prob);

		return prob;
	}

	/*determines if star with prob p should be separrated into stream*/
	public static int prob_ok(double p) {
		int ok;
		Random generator = new Random();
		double r = generator.nextDouble();
		if (p > r) {
			ok = 1;
		} else {
			ok = 0;
		}
		return ok;
	}

        /*angle between vectors a and b with range [0,pi]*/
	public static double vecangle(double[] a, double[] b) {
		double anorm = Probability.norm(a);  
		double bnorm = Probability.norm(b);
		double dprod = Probability.dotp(a, b);

		return Math.acos(dprod / (anorm * bnorm));
	}

        /* This returns the norm of a vector */
        public static double norm(double[] vector) {
                return java.lang.Math.sqrt(vector[0]*vector[0] + vector[1]*vector[1] + vector[2]*vector[2]);
        }

	public static void normalize(double[] vec) {
		double vnorm = Probability.norm(vec);
		vec[0] /= vnorm;
		vec[1] /= vnorm;
		vec[2] /= vnorm;
	}

        /* This returns the dot product of two vectors of length 3*/
        public static double dot_product(double[] a, double[] b) {
                return a[0]*b[0] + a[1]*b[1] + a[2]*b[2];
        }

	/*cross product of a and b with result put into prod*/
	public static void crossp(double[] a, double[] b, double[] prod) {
		prod[0] = a[1] * b[2] - a[2] * b[1];
		prod[1] = a[2] * b[0] - a[0] * b[2];
		prod[2] = a[0] * b[1] - a[1] * b[0];
	}

	/*dot product*/
	public static double dotp(double[] a, double[] b) {
		return a[0] * b[0] + a[1] * b[1] + a[2] * b[2];
	}

         /*calculates the stream probability that is defined by the function class StreamFunction*/
       	public static double stPsgConvolvedFunction(double[] star_coords, double[] stream_parameters, int wedge, int number_integral_points) {
		double std = .6;
		double range = 3.0*std;
		double r = star_coords[2]*1000; //in pc
		double g = Coordinates.r2mag(r);

		double upperLimit = Coordinates.mag2r(g+range);
		double lowerLimit = Coordinates.mag2r(g-range);

		StreamFunction sF = new StreamFunction();
		sF.setG(g);
		sF.setR(r);
		sF.setStd(std);
		sF.setWedge(wedge);
		sF.setStarCoords(star_coords);
		sF.setStreamParameters(stream_parameters);
		
		Integration streamInt = new Integration(sF);
		streamInt.setLimits(lowerLimit, upperLimit);
		
		double p;
		if(number_integral_points < 0) {	
			p = streamInt.trapezium(Math.abs(number_integral_points));
		} else {
			p = streamInt.gaussQuad(number_integral_points);
		}

		double prob = p/(r*r*r);

		return prob;
	}


         /*calculates the background probability that is defined by the function class BackgroundFunction*/
	public static double stPbxConvolvedFunction(double[] star_coords, double[] background_Parameters, int number_integral_points) {
                double std = .6;
                double range = 3.0*std;
                double r = star_coords[2]*1000; //in pc
		double g = Coordinates.r2mag(r);
	
                double upperLimit = Coordinates.mag2r(g+range);
                double lowerLimit = Coordinates.mag2r(g-range);

                BackgroundFunction bF = new BackgroundFunction();
                bF.setG(g);
		bF.setR(r);
		bF.setStd(std);
                bF.setStarCoords(star_coords);
                bF.setBackgroundParameters(background_Parameters);

                Integration backgroundInt = new Integration(bF);
                backgroundInt.setLimits(lowerLimit, upperLimit);

		double p;
		if (number_integral_points < 0) {
			p = backgroundInt.trapezium(Math.abs(number_integral_points));
		} else {
                	p = backgroundInt.gaussQuad(number_integral_points);
		}

		double prob = p/(r*r*r);

                return prob;
	}


	/*unconvolved stream function that applies window function to prob*/
	public static double stPsg(double[] star_coords, double[] stream_parameters, int wedge) {
		double prob = stPsgFunction(star_coords, stream_parameters, wedge);
		prob *= Coordinates.reff(star_coords[2]);
		//System.err.println("s_prob = " + prob);	
		return prob;
	}

	/*unconvolved background function that applies window function to prob*/
	public static double stPbx(double[] star_coords, double[] background_parameters) {
		double prob = stPbxFunction(star_coords, background_parameters);
		prob *= Coordinates.reff(star_coords[2]);
		//System.err.println("b_prob = " + prob);	
		
		return prob;
	}

        /*convolved stream function that applies window function to prob*/
        public static double stPsgConvolved(double[] star_coords, double[] stream_parameters, int wedge, int number_integral_points) {
                double prob = stPsgConvolvedFunction(star_coords, stream_parameters, wedge, number_integral_points);
                prob *= Coordinates.reff(star_coords[2]);

                return prob;
        }

        /*convolved background function that applies window function to prob*/
        public static double stPbxConvolved(double[] star_coords, double[] background_parameters, int number_integral_points) {
                double prob = stPbxConvolvedFunction(star_coords, background_parameters, number_integral_points);
                prob *= Coordinates.reff(star_coords[2]);

                return prob;
        }
}


/*stream probability function that consists of the stream probability convolved with a logNormal in r space*/
class StreamFunction implements IntegralFunction {
	private double g = 0.0D;
	private double r = 0.0D;
	private double std = 0.0D;
	private int wedge = 0;
	private double[] star_coords = {0.0D, 0.0D, 0.0D};
	private double[] stream_parameters = {0.0D, 0.0D, 0.0D, 0.0D, 0.0D};
	
	public double function(double x) {
		double a = 1.0/5.0;
		double b = 4.0/25.0;
		//find the value of N(x-g, std)... normal dist with mean=x-g, and std dev=std
		double denom = (x * Math.log(10) * a * std * Math.sqrt(2*java.lang.Math.PI));
		double coeff = 1 / denom;
		double logX = Math.log(x)/Math.log(10);
		double exponent = - Math.pow( (logX - (b+a*g))/(a*std*Math.sqrt(2)), 2 );
                double logN = coeff * Math.exp(exponent);

		//find value of P(x)... probability star with dist x is in stream * x^2 to account for volume		
		star_coords[2] = x/1000.0;
                double prob = Probability.stPsgFunction(star_coords, stream_parameters, wedge);
		
		double P = x*x*x*prob;

		return logN*P;
	}
	
	public void setG(double g) {
		this.g = g;
	}

        public void setR(double r) {
                this.r = r;
        }

	public void setStd(double std) {
		this.std = std;
	}
	
	public void setWedge(int wedge) {
		this.wedge = wedge;
	}

	public void setStarCoords(double[] star_coords) {
		this.star_coords[0] = star_coords[0];
		this.star_coords[1] = star_coords[1];
	}
	
	public void setStreamParameters(double[] stream_parameters) {
		for(int i = 0; i < 5; i++) {
			this.stream_parameters[i] = stream_parameters[i];
		}
	}
}
	
/*background probability function that consists of the background probability convolved with a logNormal in r space*/
class BackgroundFunction implements IntegralFunction {
	private double g = 0.0D;
	private double r = 0.0D;
	private double std = 0.0D;
        private double[] star_coords = {0.0D, 0.0D, 0.0D};
        private double[] background_parameters = {0.0D, 0.0D, 0.0D, 0.0D};

	
	public double function(double x) {
                double a = 1.0/5.0;
                double b = 4.0/25.0;
                //find the value of logN(x-r', std)... normal dist with mean=x-g, and std dev=std
                double denom = (x * Math.log(10) * a * std * Math.sqrt(2*java.lang.Math.PI));
                double coeff = 1 / denom;
                double logX = Math.log(x)/Math.log(10);
                double exponent = - Math.pow( (logX - (b+a*g))/(a*std*Math.sqrt(2)), 2 );
                double logN = coeff * Math.exp(exponent);
                
		//find value of P(x)... probability star with dist x is in background * x^2 to account for volume 
                star_coords[2] = x/1000.0;
                double prob = Probability.stPbxFunction(star_coords, background_parameters);
		double P = x*x*x*prob;
                return logN*P;
	}
	
	public void setG(double g) {
		this.g = g;
	}

        public void setR(double r) {
                this.r = r;
        }
		
	public void setStd(double std) {
		this.std = std;
	}

        public void setStarCoords(double[] star_coords) {
                this.star_coords[0] = star_coords[0];
                this.star_coords[1] = star_coords[1];
        }

        public void setBackgroundParameters(double[] background_parameters) {
		for(int i = 0; i < 4; i++) {
			this.background_parameters[i] = background_parameters[i];
		}
	}

}
